(*

KodeZwerg's Custom Control Template
Offering: visible control with colors, text and border, clickable
Extra: autosize and angle setting (it can rotate around its center)
License: The Unlicense

Notes:
This template is fully based on the LCL and Graphics32 library, flicker free all the time due doublebuffer

Known bugs:
- Positioning is incorrect when autosize is activated during angle change
- increasing bordersize too much will make graphic appear strange

Future plans:
- rewrite component to use Bitmap32 all the time, including its methods and just at painting on screen assign to regular TBitmap
- rewrite OnClick event to properly know if clicked inside painted rectangle, currently its done by simple compare colors
- when OnClick works like above line told, implement Glyph feature to draw a small Bitmap on Control
- maybe add shadow or gradient effects...
*)

unit kz.CustomControl;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, Forms, Controls, Graphics,
  LCLType, LCLIntf,
//  Math, GraphMath,
  GR32, GR32_Transforms;

type

  { TkzCustomControl }

  TkzCustomControl = class(TCustomControl)
    strict private
      FBMP: TBitmap;
      FText: string;
      FAngle,
      FVirtualWidth,
      FVirtualHeight,
      FCenterX,
      FCenterY,
      FBorderSize: Integer;
      FUpdate,
      FAutoSize: Boolean;
      FColor,
      FBorderColor: TColor;
      FFont: TFont;
      FOnClick: TNotifyEvent;
    private
      procedure SetAngle(const AValue: Integer);
      procedure SetAutoSize(const AValue: Boolean); reintroduce;
      procedure SetBorderColor(const AValue: TColor);
      procedure SetBorderSize(const AValue: Integer);
      procedure SetCaption(const AValue: string);
      procedure SetColor(const AValue: TColor); reintroduce;
      procedure SetFont(const AValue: TFont);
    protected
      procedure Paint; override;
      procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    public
      constructor Create(const AOwner: TComponent; const AParent: TWinControl); reintroduce;
      destructor Destroy; override;
      procedure Update; reintroduce;
      procedure Repaint; reintroduce;
      procedure Refresh;
    published
      // Background Color
      property Color: TColor read FColor write SetColor default clBtnFace;
      // Font, initialized with whatever your OS default font is, including settings
      property Font: TFont read FFont write SetFont;
      // Text on Control
      property Caption: string read FText write SetCaption;
      // Width of Border, 0 to turn off, Default 1
      property BorderSize: Integer read FBorderSize write SetBorderSize default 1;
      // Color of Border
      property BorderColor: TColor read FBorderColor write SetBorderColor default clWindowFrame;
      // Angle of Control, range: 0° - 360°
      property Angle: Integer read FAngle write SetAngle default 0;
      // Control use Caption to autosize itself
      property AutoSize: Boolean read FAutoSize write SetAutoSize default False;
      // basic click event
      property OnClick: TNotifyEvent read FOnClick write FOnClick;
  end;

implementation

constructor TkzCustomControl.Create(const AOwner: TComponent; const AParent: TWinControl);
var
  FontData: TFontData;
begin
  inherited Create(AOwner);
  Self.Parent := AParent;
  Self.Visible := False;
  FOnClick := nil;
  FAngle := 0;
  FText := 'KodeZwergs Custom Control';
  FBMP := nil;
  FVirtualWidth := 0;
  FVirtualHeight := 0;
  FCenterX := 0;
  FCenterY := 0;
  FBorderSize := 1;
  FColor := clBtnFace;
  FBorderColor := clWindowFrame;
  FFont := TFont.Create;
  try
    FontData := GetFontData(Screen.SystemFont.Handle);
    FFont.Name := FontData.Name;
    FFont.Size := Round(FontData.Height * 72 / Screen.PixelsPerInch);
    FFont.Orientation := FontData.Orientation;
    FFont.CharSet := FontData.CharSet;
    FFont.Pitch := FontData.Pitch;
    FFont.Quality := FontData.Quality;
    FFont.Style := FontData.Style;
    FFont.Color := clBtnText;
  finally
  end;
  FUpdate := False;
  FAutoSize := False;
end;

destructor TkzCustomControl.Destroy;
begin
  if (FBMP <> nil) then
    FBMP.Free;
  FFont.Free;
  FOnClick := nil;
  inherited Destroy;
end;

procedure TkzCustomControl.Update;
begin
  FUpdate := True;
  Self.Invalidate;
end;

procedure TkzCustomControl.Repaint;
begin
  FUpdate := True;
  Self.Invalidate;
end;

procedure TkzCustomControl.Refresh;
begin
  FUpdate := True;
  Self.Invalidate;
end;

procedure TkzCustomControl.SetFont(const AValue: TFont);
begin
  FFont := AValue;
  FUpdate := True;
  Self.Invalidate;
end;

procedure TkzCustomControl.SetCaption(const AValue: string);
begin
  if (FText = AValue) then
    Exit;
  FText := AValue;
  FUpdate := True;
  Self.Invalidate;
end;

procedure TkzCustomControl.SetAutoSize(const AValue: Boolean);
begin
  if (FAutoSize = AValue) then
    Exit;
  FAutoSize := AValue;
  if FAutoSize then
    begin
      FUpdate := True;
      Self.Invalidate;
    end;
end;

procedure TkzCustomControl.SetBorderSize(const AValue: Integer);
begin
  if ((FBorderSize = AValue) or (AValue < 0)) then
    Exit;
  FBorderSize := AValue;
  FUpdate := True;
  Self.Invalidate;
end;

procedure TkzCustomControl.SetColor(const AValue: TColor);
begin
  if (AValue = FColor) then
    Exit;
  FColor := AValue;
  FUpdate := True;
  Self.Invalidate;
end;

procedure TkzCustomControl.SetAngle(const AValue: Integer);
begin
  if (FAngle = AValue) then
    Exit;
  if ((AValue < 0) or (AValue > 360)) then
    Exit;
  FAngle := AValue;
  FUpdate := True;
  Self.Invalidate;
end;

procedure TkzCustomControl.SetBorderColor(const AValue: TColor);
begin
  if (FBorderColor = AValue) then
    Exit;
  FBorderColor := AValue;
  FUpdate := True;
  Self.Invalidate;
end;

procedure TkzCustomControl.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
//  if (Self.Canvas.Pixels[X, Y] <> Self.Parent.Color) then
  if ((Self.Canvas.Pixels[X, Y] = FColor) or (Self.Canvas.Pixels[X, Y] = FBorderColor) or (Self.Canvas.Pixels[X, Y] = FFont.Color)) then
    begin
      if Assigned(FOnClick) then
        FOnClick(Self);
    end;
end;

// https://stackoverflow.com/questions/10633400/rotate-bitmap-by-real-angle
// method by NGLN

procedure RotateBitmapGR32(Bmp: TBitmap32; Degs: Integer; AdjustSize: Boolean;
  BkColor: TColor = clNone; Transparent: Boolean = False);
var
  Tmp: TBitmap32;
  Transformation: TAffineTransformation;
begin
  Tmp := TBitmap32.Create;
  Transformation := TAffineTransformation.Create;
  try
    Transformation.BeginUpdate;
    Transformation.SrcRect := FloatRect(0, 0, Bmp.Width, Bmp.Height);
    Transformation.Translate(-0.5 * Bmp.Width, -0.5 * Bmp.Height);
    Transformation.Rotate(0, 0, -Degs);
    if AdjustSize then
      with Transformation.GetTransformedBounds do
        Tmp.SetSize(Round(Right - Left), Round(Bottom - Top))
    else
      Tmp.SetSize(Bmp.Width, Bmp.Height);
    Transformation.Translate(0.5 * Tmp.Width, 0.5 * Tmp.Height);
    Transformation.EndUpdate;
    Tmp.Clear(Color32(BkColor));
    if not Transparent then
      Bmp.DrawMode := dmOpaque
    else
      Bmp.DrawMode := dmTransparent;
    Transform(Tmp, Bmp, Transformation);
    Bmp.Assign(Tmp);
    Bmp.OuterColor := Color32(BkColor);
  finally
    Transformation.Free;
    Tmp.Free;
  end;
end;

procedure RotateBitmap(Bmp: TBitmap; Degs: Integer; AdjustSize: Boolean;
  BkColor: TColor = clNone);
var
  Tmp: TBitmap32;
  Transparent: Boolean;
begin
  Tmp := TBitmap32.Create;
  try
    Transparent := Bmp.Transparent;
    Tmp.Assign(Bmp);
    RotateBitmapGR32(Tmp, Degs, AdjustSize, BkColor, Transparent);
    Bmp.Assign(Tmp);
    if Transparent then
      Bmp.Transparent := True;
  finally
    Tmp.Free;
  end;
end;

procedure TkzCustomControl.Paint;
var
  LBMP: TBitmap;
  R: TRect;
  OffsetX, OffsetY: Integer;
begin
  // create the root image only when needed
  if ((FBMP = nil) or FUpdate) then
    begin
      if (FBMP= nil) then
        FBMP := TBitmap.Create;
      try
        FBMP.Canvas.AntialiasingMode := amOn;
        FBMP.Canvas.Font := FFont;
        if FAutoSize then
          begin
            Self.Width := FBMP.Canvas.TextWidth(FText) + 1 + FBorderSize;
            Self.Height := FBMP.Canvas.TextHeight(FText) + 1 + FBorderSize;
            FVirtualWidth := Self.Width;
            FVirtualHeight := Self.Height;
          end;
        if (FVirtualWidth = 0) then
          FVirtualWidth := Self.Width;
        if (FVirtualHeight = 0) then
          FVirtualHeight := Self.Height;
        if (FCenterX = 0) then
          FCenterX := Self.Left + (FVirtualWidth div 2);
        if (FCenterY = 0) then
          FCenterY := Self.Top + (FVirtualHeight div 2);
        FBMP.SetSize(FVirtualWidth, FVirtualHeight);
        FBMP.Canvas.Brush.Style := bsSolid;
        FBMP.Canvas.Brush.Color := FColor;
        FBMP.Canvas.FillRect(Rect(0, 0, FVirtualWidth, FVirtualHeight));
        if FBorderSize > 0 then
          begin
            FBMP.Canvas.Pen.Style := psSolid;
            FBMP.Canvas.Pen.Color := FBorderColor;
            FBMP.Canvas.Pen.Width := FBorderSize;
            FBMP.Canvas.Rectangle(0, 0, FVirtualWidth, FVirtualHeight);
          end;
        FBMP.Canvas.Pen.Color := Self.Font.Color;
        R := Rect(FBorderSize, FBorderSize, FVirtualWidth - FBorderSize, FVirtualHeight - FBorderSize);
        FBMP.Canvas.Pen.Color := FFont.Color;
        DrawText(FBMP.Canvas.Handle, PChar(FText), Length(FText), R, DT_CENTER or DT_VCENTER or DT_SINGLELINE);
      finally
        FUpdate := False;
      end;
    end;

  LBMP := TBitmap.Create;
  try
    LBMP.Assign(FBMP);

    RotateBitmap(LBMP, FAngle, True, clDefault);

    // set new dimension for control
    Self.Width := LBMP.Width;
    Self.Height := LBMP.Height;

    // Calculate the offset between the old center and the new center
    OffsetX := FCenterX - (Self.Width div 2);
    OffsetY := FCenterY - (Self.Height div 2);

    // Adjust the position of the bitmap
    Self.Left := OffsetX;
    Self.Top := OffsetY;

    // Draw the rotated bitmap onto the control's canvas
    Self.Canvas.Draw(0, 0, LBMP);
  finally
    LBMP.Free;
  end;
end;


end.

